/***************************************************************************
* Multiline.C															   *
*																		   *
* The Multiline class encapsulates sequences of 2D points with time-	   *
* stamps; the sequences encode one or more "polylines", which are chains   *
* of line segments.  The class automatically maintains information such    *
* as the bounding box, length, number of disconnected segments, etc.	   *
*																		   *
*	HISTORY 															   *
*	  10/09/99	  arvo	  Added Resample function.						   *
*	  08/20/99	  arvo	  Replaced Read and Write with stream operators.   *
*	  08/09/98	  arvo	  Added Distance, Split, and Splice.			   *
*	  05/14/98	  arvo	  Converted Polyline to Multiline.				   *
*	  05/12/98	  arvo	  Split off into a separate module. 			   *
*	  04/04/98	  arvo	  Added Write method to Multiline class.		   *
*	  12/09/97	  arvo	  Added Read method to Multiline class. 		   *
*	  12/04/97	  arvo	  Reorganized for use as a library. 			   *
*	  07/07/97	  arvo	  Added TimedVec2 class.						   *
*	  07/03/97	  arvo	  Multilines can now encode jumps.				   *
*	  06/17/97	  arvo	  Added multiline code. 						   *
*	  07/09/96	  arvo	  Initial coding.								   *
*																		   *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo										   *
*																		   *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html	   *
*																		   *
* This program is distributed in the hope that it will be useful, but	   *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more	   *
* details.																   *
*																		   *
***************************************************************************/
#include "stdafx.h"

#include <assert.h>
#include <stdio.h>
#include <fstream>
using namespace std;
#include "../util/Char.h"
#include "../util/Multiline.h"

extern ofstream cerr, cout;

Pnode*			Multiline::free_node = NULL;
long			Multiline::num_nodes = 0;
long			Multiline::num_lines = 0;
const Vec2		Multiline::Jump;
const Multiline Multiline::Null;

Vec2 Interpolate( const Pnode *A, const Pnode *B, double s )
{
	// Compute a point along the segment joining Pnodes A and B
	// according to the arlength parameter s.
	//	
	double d = B->Param() - A->Param();
	if( d > 0.0 )
	{
		double t = ( s - A->Param() ) / d;
		return Interpolate( A->Coord(), B->Coord(), t );
	}
	return A->Coord();
}

/***************************************************************************
*																		   *
* Multiline:  A sequence of 2D points representing one or more sequences   *
*			 of connected lines.										   *
*																		   *
***************************************************************************/

Multiline::Multiline()
{
	first = NULL; // Initialize first and last pointers so that
	last  = NULL; // the clear method does no garbage collection.
	Clear();	  // This initializes all other fields.
	num_lines++;  // Count all currently allocated multilines.
}

Multiline::Multiline( const Multiline &P ) 
{ 
	// Copy constructor based on the assignment operator.
	first = NULL; 
	last  = NULL;
	*this = P; 
	num_lines++; // Count all currently allocated multilines.
}

Multiline::Multiline( const Multiline *P ) 
{ 
	// Copy constructor based on the assignment operator.
	first = NULL; 
	last  = NULL;
	if( P != NULL ) *this = *P; 
	num_lines++; // Count all currently allocated multilines.
}

Multiline::Multiline( const Box2d &B ) 
{ 
	first = NULL; 
	last  = NULL;
	*this = B; 
	num_lines++; // Count all currently allocated multilines.
}

Multiline::~Multiline()
{
	Clear();	 // Destroy all vertices & garbage collect them.
	num_lines--; // Update counter for all currently allocated multilines.
}

Pnode *Multiline::NewNode( const Vec2 &coord )
{
	// Allocate a new vertex node from the internal free list.
	// Because these nodes are small, of uniform size, and
	// already chained together, it is more efficient to do the
	// memory management with an explicit free list than to
	// use the built-in C++ memory management.
	Pnode *p;
	if( free_node == NULL ) 
	{
		p = new Pnode;
		num_nodes++;
	}
	else
	{
		p = free_node;
		free_node = free_node->next;
	}
	p->Coord() = coord;
	p->param   = 0.0;
	p->next    = NULL;
	p->time    = 0;
	return p;
}

void Multiline::Free( Pnode *from, Pnode *to )
{
	// Garbage collect an entire list of Pnodes by splicing them
	// into the free list.	Here "from" must be the first node
	// in the list, and "to" must be the last node in the list.
	// This way, we needn't search for the end of the list.
	if( from != NULL )
	{
		assert( to != NULL );
		to->next  = free_node;
		free_node = from; 
	}
}

void Multiline::Clear()
{
	// Garbage collect the vertices and re-initialize all fields.
	// First, splice the linked-list of vertices onto the head of
	// the free list (if there are any).
	Free( first, last );
	
	// Initialize all the other fields.
	first  = NULL;
	last   = NULL;
	penult = NULL;
	box    = Box2d::Null;
	length = 0.0;
	seglen = 0.0;
	points = 0;
	comps  = 0;
	jump   = 0;
	dirty  = 0;
	weight = 1;
}

Box2d Multiline::Box() const  // Box containing all vertices, expanded by line weight.
{
	if( weight > 1 ) return box.Expand( weight );
	return box;
}

Box2d Multiline::TrueBox() // Box containing all vertices, expanded by line weight.
{
	if( dirty ) RecomputeBox();
	return Box();
}

Vec2 Multiline::Start() const
{ 
	// Return the first vertex, if there is one.
	if( First() == NULL ) return Vec2::Zero;
	else return *First();
}

Vec2 Multiline::End() const
{ 
	// Return the last vertex (of the last segment), if there is one.
	if( Last() == NULL ) return Vec2::Zero;
	else return *Last();
}

long Multiline::Started() const
{ 
	// Return the time-stamp of the first vertex, if there is one.
	if( First() == NULL ) return 0;
	else return First()->Time();
}

long Multiline::Ended() const
{ 
	// Return the time-stamp of the last vertex of the last segment.
	if( Last() == NULL ) return 0;
	else return Last()->Time();
}

Vec2 Multiline::StartOf( int index ) const 
{
	// Return coords of first vertex in the given (zero-based) component.
	Pnode *n = CompPtr( index );
	if( n == NULL ) return Vec2::Zero;
	else return n->Coord();
}

Vec2 Multiline::EndOf( int index ) const 
{
	// Return coords of last vertex of the given component.
	Pnode *n = CompPtr( index );
	if( n != NULL ) 
	{
		while( n->Next() != NULL && !n->Next()->Jump() )
			n = n->Next();
		return n->Coord();
	}
	return Vec2::Zero;
}

double Multiline::LengthOf( int index ) const 
{
	// Return length of the given component.
	Pnode *n = CompPtr( index );
	if( n != NULL ) 
	{
		while( n->Next() != NULL && !n->Next()->Jump() )
			n = n->Next();
		return n->Param();
	}
	return 0.0;
}

double Multiline::AreaOf( int index ) const 
{
	// Return signed area of the given component.
	Multiline Temp;
	Pnode *n = CompPtr( index );
	if( n == NULL ) return 0.0;
	do	{
		Temp << *n;
		n = n->Next();
	}
	while( n != NULL && !n->Jump() );
	return SignedArea( Temp );
}

Pnode *Multiline::CompPtr( int index ) const 
{
	// Return pointer to first node in the given (zero-based) component.
	if( index >= Comps() ) return NULL;
	for( Pnode *n = First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() )
		{
			if( index <= 0 ) return n;
			index--;
		}
	}
	return NULL;
}

Vec2 Multiline::Vertex( int n ) const // Coords of vertex (zero-based).
{
	int count = 0;
	if( Points() <= n ) return End();
	for( Pnode *v = First(); v != NULL; v = v->Next() )
	{
		if( count++ == n ) return *v;
	}
	return End();
}

Vec2 Multiline::Closest( const Vec2 &P, Pnode* &node, double &param ) const
{
	// Compute the point on the multiline closest to the point P.
	// Also return the pointer to the Pnode corresponding to the
	// first endpoint of the segment containing the closest point.
	// That is, the returned point will lie on the segment connecting
	// node->Coord() and node->Next()->Coord().
	double w;
	double dist2 = Infinity;
	Vec2  Q, A, X;
	Pnode *prev = NULL;
	for( Pnode *curr = First(); curr != NULL; curr = curr->Next() )
	{
		Vec2 B = curr->Coord();
		if( !curr->Jump() )
		{
			Vec2 R = P - A;
			Vec2 V = B - A;
			double s = R * V;
			double t = LenSqr( V );
			if( s <= 0.0 ) // The point A is closest.
			{
				w = prev->Param();
				X = A;
			}
			else if( s >= t ) // The point B is closest.
			{
				w = curr->Param();
				X = B;
			}
			else // The closest point is in the interior.
			{
				double r = s / t;
				w = ( 1.0 - r ) * prev->Param() + r * curr->Param();
				X = A + r * V;
			}
			double d2 = LenSqr( P - X );
			if( d2 < dist2 )
			{
				dist2 = d2;   // The distance squared.
				Q	  = X;	  // The coords of the point.
				param = w;	  // Arclength parameter of Q;
				node  = prev; // An endpoint of the segment.
			}
		}
		A = B;
		prev = curr;
	}
	return Q;
}

Vec2 Multiline::Closest( const Vec2 &P ) const
{
	Pnode *pnode;
	double param;
	return Closest( P, pnode, param );
}

Vec2 Multiline::Closest( const Vec2 &P, Pnode* &pnode ) const
{
	double param;
	return Closest( P, pnode, param );
}

Vec2 Multiline::Closest( const Vec2 &P, double &param ) const
{
	Pnode *pnode;
	return Closest( P, pnode, param );
}

Multiline &Multiline::operator<<( const Vec2 &p )  // Add a point to the list.
{
	// The << operator adds a point to the end of the multiline.  This is
	// the standard method of constructing multilines.	Jumps are encoded
	// by "inserting" a special point Multiline::Jump, which is simply used
	// as a marker (i.e. it is not literally inserted).
	//--------------------------------------------------------------------
	// First check for the special case of the point Multiline::Jump
	// being inserted.	This is not a real point, but a marker to begin
	// a new segment.  The beginning of a new segment in an existing
	// Multiline is indicated by a parameter of zero.
	if( &p == &Jump )
	{
		if( jump == 1 ) return *this; // Ignore repeated jump markers.
		jump = 1;					  // Flag the next point as a jump.
		return *this;				  // Do NOT actually insert the point.
	}
	box << p; // This is a legitimate point, so expand the bounding box.
	Pnode *n = NewNode( p );
	if( first == NULL ) // The very first point.
	{
		comps	 = 1;
		first	 = n;
		length	 = 0.0;
		seglen	 = 0.0;
		n->param = 0.0;
	}
	else if( jump ) // Begin a new component.
	{
		comps++; // Bump component counter at first real point after a jump.
		seglen	   = 0;
		n->param   = 0.0;
		last->next = n;
	}
	else // Add the point to an existing segment.
	{
		double step = dist( *last, p );
		length += step;
		seglen += step;
		last->next = n;
		n->param = seglen;
	}
	points++;
	n->next = NULL;
	penult	= last;
	last	= n;
	jump	= 0;
	return *this;
}

Multiline &Multiline::operator<<( const TimedVec2 &p )	
{
	// Add a point with a time stamp.
	*this << p.Coord();
	Last()->time = p.Time();
	return *this;
}

Multiline &Multiline::operator+=( const Vec2 &p )  
{
	// Translate this multiline by translating each of the vertices
	// and the bounding box.
	for( Pnode *n = First(); n != NULL; n = n->Next() )
	{
		n->Coord() += p;
	}
	box += p;  // Translate the bounding box.
	return *this;
}

Multiline &Multiline::operator-=( const Vec2 &p )
{
	// Translate this multiline by translating each of the vertices
	// and the bounding box.
	for( Pnode *n = First(); n != NULL; n = n->Next() )
	{
		n->Coord() -= p;
	}
	box -= p;  // Translate the bounding box.
	return *this;
}

Multiline &Multiline::operator*=( double scale ) 
{
	// Scale this multiline by scaling each of the vertices.
	// Also scale the bounding box and the length.
	for( register Pnode *n = First(); n != NULL; n = n->Next() )
	{
		n->Coord() *= scale;
		n->param   *= scale;
	}
	length *= scale;  // Scale the line length.
	box    *= scale;  // Scale the bounding box.
	return *this;
}

Multiline &Multiline::operator*=( const Mat2x2 &M ) 
{
	// Transform all vertices.
	Multiline Temp;
	for( register Pnode *n = First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() )  Temp << Multiline::Jump;
		Temp << M * n->Coord();
	}
	return *this = Temp;
}

Multiline &Multiline::operator*=( const Trans2 &T ) 
{
	// Transform all vertices.
	Multiline Temp;
	for( register Pnode *n = First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() )  Temp << Multiline::Jump;
		Temp << T * n->Coord();
	}
	return *this = Temp;
}

Multiline &Multiline::operator/=( const Trans2 &T ) 
{
	// Transform all vertices.
	Multiline Temp;
	for( register Pnode *n = First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() )  Temp << Multiline::Jump;
		Temp << n->Coord() / T;
	}
	return *this = Temp;
}

Multiline &Multiline::operator^=( double scale ) 
{
	// Scale this multiline with respect to its center by performing
	// the corresponding operation on the vertices and the bounding box.
	Vec2 C = Center();
	for( Pnode *n = First(); n != NULL; n = n->Next() )
	{
		n->Coord() = scale * ( n->Coord() - C ) + C;
		n->param  *= scale;
	}
	length *= scale;  // Scale the line length.
	box    ^= scale;  // Scale the bounding box about its center.
	return *this;
}

Multiline &Multiline::operator=( const Multiline &P )
{
	// The copy constructor simply clears the current multiline and
	// then rebuilds it, one vertex at a time.	Jumps must be recovered
	// by looking for parameters of zero.
	Clear();
	for( const Pnode *n = P.First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() ) *this << Multiline::Jump;
		*this << *n;
	}
	weight = P.Weight();
	return *this;
}

Multiline &Multiline::operator+=( const Multiline &P ) 
{
	// Add the given multiline to the end of this one, after a jump.
	// Since each jump in P is copied as well (including the first),
	// all components of P will be appended to this multiline.
	*this << Multiline::Jump; 
	for( const Pnode *n = P.First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() ) *this << Multiline::Jump;
		*this << *n;
	}
	return *this;
}	

Multiline &Multiline::operator<<( const Multiline &P ) 
{
	// Append the given multiline to the end of this one, with NO jump.
	// Subsequent jumps in P will be copied, but the first component of
	// P will be connected to the last component of the current multiline.
	for( const Pnode *n = P.First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() && n != P.First() ) *this << Multiline::Jump;
		*this << *n;
	}
	return *this;
}	

Multiline &Multiline::operator*=( const Perm &perm )
{
	// Permute the components of the given multiline by the given 
	// permutation.  The permutation must have the same number of 
	// elements as the multiline has components.
	int n = Comps();
	if( n != perm.Size() )
	{
		cerr << "ERROR: in Multiline * Perm operator" << endl;
		return *this;
	}
	
	// Create an array of multilines corresponding to the components of
	// this multiline, and blow away the original version.
	
	Multiline *comps = Explode( *this );
	Clear();
	
	// Copy the permuted components of this multiline back into itself.
	for( int j = 0; j < n; j++ ) *this += comps[ perm[j] ];
	
	delete[] comps; // Dispose of the individual components.
	return *this;
}

Vec2 Multiline::operator()( double u ) const
{
	// Compute the point that corresponds to parameter u in [0,1], 
	// where 0 is the first vertex and 1 is the last vertex of the LAST 
	// component.
	double t = Clamp( 0.0, u, 1.0 );
	return PointAt( t * Length() );
}

Vec2 Multiline::PointAt( double s, Vec2 *tangent ) const
{
	// Compute the point that corresponds to parameter u in [0,1], 
	// where 0 is the first vertex and 1 is the last vertex of the LAST 
	// component.
	if( s <= 0.0	  ) return Start();
	if( s >= Length() ) return End();
	double dist = 0;
	double seg  = 0;
	const Pnode *p = NULL;
	for( const Pnode *q = First(); q != NULL; q = q->Next() )
	{
		// Accumulate the total distance so that this will work with
		// multiple components.
		if( q->Jump() ) seg = dist;
		dist = seg + q->param;
		if( dist >= s && p != NULL ) 
		{
			if( tangent != NULL ) *tangent = *q - *p;
			return Interpolate( p, q, s );
		}
		p = q;
	}		 
	if( tangent != NULL && Penult() != NULL ) *tangent = End() - *Penult();
	return End();
}

Multiline &Multiline::operator=( const Box2d &B )  
{
	// Make a multiline from a box.
	Clear();
	*this << B.NE() << B.NW() << B.SW() << B.SE() << B.NE();
	return *this;
}

Multiline &Multiline::Reverse()
{
	// Destructively reverse the given multiline and update its
	// parametrization.  (This currently does not handle jumps!)
	// First reverse the directions of all the pointers, then
	// reset the first, penult, and last pointers.
	//
	Pnode *prev = NULL;
	Pnode *n	= First();
	while( n != NULL )
	{
		Pnode *next = n->Next();
		n->next = prev;
		prev	= n;
		n		= next;
	}
	last   = First();
	penult = First()->Next();
	first  = prev != NULL ? prev : last;
	
	// Now recompute the parametrization.  
	// The length does not need to be changed.
	
	First()->param = 0.0;
	double	len    = 0.0;
	prev = First();
	for( Pnode *p = First()->next; p != NULL; p = p->Next() )
	{
		len += dist( prev->Coord(), p->Coord() );
		p->param = len;
		prev = p;
	}
	return *this;
}

Multiline &Multiline::Pop( Multiline &P )
{
	// This operator removes the first component (i.e. connected sequence)
	// from this multiline, and replaces the multiline on the right with it.
	// The original multiline (with the first component removed) is returned
	// as the value of this function.
	//
	P.Clear(); // Prepare the multiline that will hold the first component.
	
	Pnode *prev = NULL; // Alwyas points to previous Pnode.
	
	for( Pnode *n = First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() && prev != NULL )
		{
			length -= prev->Param();
			points -= P.Points();
			comps--;
			dirty = 1;	// The bounding box may be too large now.
			Free( first, prev );
			first = n;
			RecomputeBox();
			return *this;
		}
		prev = n;
		P << *n;
	}
	Clear(); // This multiline was one connected sequence.
	return *this;
}

Multiline Multiline::Pop( )
{
	Multiline M; 
	Pnode *prev = NULL; // Alwyas points to previous Pnode.
	
	for( Pnode *n = First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() && prev != NULL )
		{
			length -= prev->Param();
			points -= M.Points();
			comps--;
			dirty = 1;	// The bounding box may be too large now.
			Free( first, prev );
			first = n;
			RecomputeBox();
			return M;
		}
		prev = n;
		M << *n;
	}
	Clear(); // This multiline was one connected sequence.
	return M;
}

Vec2 Multiline::Move( int n, const Vec2 &delta )
{
	// Locate the n'th vertex, change it, and then fix up
	// all the arclength parameters that follow, until the
	// end of the component.
	
	if( Points() <= n ) return Vec2::Zero;
	Pnode *u = NULL;
	Pnode *v = First();
	
	// Find the n'th vertex, and the one before it (i.e. n-1).
	for( int i = 0; i < n; i++ )
	{
		u = v;
		v = v->Next();
		assert( v != NULL );
	}
	
	double change = 0.0;
	v->Coord() += delta;
	Vec2 new_coord = v->Coord();
	// Expand the bounding box if necessary.
	box << v->Coord();
	if( u != NULL && !v->Jump() ) 
	{
		double new_param = u->Param() + dist( *u, *v );
		change = new_param - v->Param();
		v->param = new_param;
	}
	
	u = v;
	v = v->Next();
	if( v != NULL && !v->Jump() ) 
	{
		double new_param = u->Param() + dist( *u, *v );
		change = new_param - v->Param();
		v->param = new_param;
	}
	
	if( v != NULL && !v->Jump() )
	{
		v = v->Next();
		while( v != NULL && !v->Jump() )
		{
			v->param += change;
			v = v->Next();
		}
	}
	// Adjust the overall length, and flag the bounding
	// box as dirty.
	length += change;
	dirty = 1;
	return new_coord;
}

void Multiline::RecomputeBox()
{
	if( dirty )
	{
		box = Box2d::Null;
		for( Pnode *v = First(); v != NULL; v = v->Next() ) box << *v;
		dirty = 0;
	}
}

class Inode : public Vec2 {
public:
	double delta;
	int   left;
};

/*-------------------------------------------------------------------------*
*  I N S I D E															  *
*																		  *
*  Determines whether a 2D point is within the interior of a polygon.	  *
*																		  *
*-------------------------------------------------------------------------*/
int Multiline::Inside( const Vec2 &P ) const
{
	Inode A, B, C;
	int   cross = 0;
	
	// Loop setup.	Define "previous" values of A and B.
	
	Pnode *v = Last();
	A.X()	 = v->X();
	A.Y()	 = v->Y();
	A.left	 = A.X() <= P.X();
	A.delta  = P.Y() - A.Y();
	B.delta  = P.Y() - Penult()->Y();
	
	for( v = First(); v != NULL; v = v->Next() )
	{
		// Copy point A into point B, and all information concerning point A.
		// All we need to retain from B is the "delta".
		
		C.delta = B.delta;
		B = A;
		
		// Get the next point A and fill in its positional information.
		
		A.X()	= v->X();
		A.Y()	= v->Y();
		A.left	= A.X() <= P.X();
		A.delta = P.Y() - A.Y();
		
		// Skip to next point if segment AB is above or below line thru P.
		
		if( A.delta > 0 && B.delta > 0 ) continue;
		if( A.delta < 0 && B.delta < 0 ) continue;
		
		// Skip to next point if neither A nor B is to the right of P. 
		
		if( A.left && B.left ) continue;
		
		// If B is on the line and to the right of P, we must decide whether
		// to count it as a crossing by looking at the positions of A and C.
		
		if( B.delta == 0 )
		{
			if( B.left ) continue;
			if( ( A.delta > 0 && C.delta <= 0 ) ||
				( C.delta > 0 && A.delta <= 0 ) ) cross++;
			continue;
		}
		
		// If A is on the line we can just skip to the next line segment since 
		// we are assuming that the segment AB is open at A.  We will deal with
		// this point the next time around when it becomes B.
		
		if( A.delta == 0 ) continue;
		
		// Assertion: A & B should be on opposite sides of the horizontal line 
		// thu P with at least one of them to the right of P.  If both A and B
		// are to the right of P there is a crossing.  If one point is to the
		// left of P we must compute the actual intersection point in order to
		// determine which side of P it is on.
		
		if( A.left || B.left )
		{
			double t1 = B.Y() - A.Y();
			double t2 = ( P.Y() - A.Y() ) * ( B.X() - A.X() );
			double t3 = ( P.X() - A.X() ) * Abs( t1 );
			double t4 = ( t1 > 0 ) ? t2 : -t2;
			if( t3 < t4 ) cross++;
		}
		else cross++;  // Both A and B are to the right.
	}
	
	// If the number of crossings is odd, the point is in the interior or on
	// the boundary.  If it is even, the point is exterior or on the boundary.
	
	return Odd( cross );
}

/***************************************************************************
*																		   *
* Non-Member Multiline Functions										   *
*																		   *
***************************************************************************/

Multiline operator*( const Trans2 &trans, const Multiline &lines )
{
	Multiline M;
	for( Pnode *n = lines.First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() ) M << Multiline::Jump;
		M << trans * n->Coord();
	}
	return M;
}

double GapLength( const Multiline &M )
{
	// Returns the lengh of all jumps combined.
	if( M.Comps() <= 1 ) return 0.0;
	double gaplen = 0.0;
	Pnode *prev  = M.First();
	for( Pnode *n = prev->Next(); n != NULL; n = n->Next() )
	{
		if( n->Jump() ) gaplen += dist( prev->Coord(), n->Coord() );
		prev = n;
	}
	return gaplen;
}

Multiline Truncate( const Multiline &M, double s )
{
	Multiline Q;
	Pnode *p = M.First();
	Pnode *q = p->Next();
	Q << *p;
	while( q->Param() < s )
	{
		Q << *q;
		p = q;
		q = q->Next();
		if( q == NULL ) return Q;
	}
	Q << Interpolate( p, q, s );
	return Q;
}

void Split( const Multiline &M, double s, Multiline &A, Multiline &B )
{
	// Generate two new multilines from M by splitting it at a point
	// which is the given arclength "s" along the line.  This works
	// for multilines with any number of segments.
	A.Clear();
	B.Clear();
	Pnode *p = M.First();
	Pnode *q = p->Next();
	Vec2 split_point;
	
	// Copy the the part of M from 0 to s into A.
	A << *p;
	while( q != NULL )
	{
		if( q->Param() >= s )
		{
			split_point = Interpolate( p, q, s );
			A << split_point;
			break;
		}
		if( q->Jump() ) A << Multiline::Jump;
		A << *q;
		p = q;
		q = q->Next();
	}
	if( q == NULL ) return;
	
	// Now copy the rest of the polyline into B, starting
	// from the split point.
	if( s != q->Param() ) B << split_point;
	while( q != NULL )
	{
		if( q->Jump() ) A << Multiline::Jump;
		B << *q;
		q = q->Next();
	}
}

Multiline CopyPointsSince( const Multiline &lines, long start )
{
	Multiline M;
	Pnode *n = lines.First();
	
	// Skip to first vertex that has a timestamp after "start" time.
	while( n != NULL && n->Time() < start ) n = n->Next();
	
	// Now add the current vertex and all remaining vertices to this multiline.
	while( n != NULL )
	{
		if( n->Jump() ) M << Multiline::Jump;
		M << *n;
		n = n->Next();
	}
	return M;
}

Multiline Partition( const Multiline &M, int n, double split[] )
{
	// This currently only works for single-segment multilines!
	const double epsilon = 1.0E-5; 
	Multiline P;
	int   k = 0;
	double s = Abs( split[0] * M.Length() );
	double r = 0.0;
	Pnode *p = NULL;
	for( Pnode *q = M.First(); q != NULL; q = q->Next() )
	{
		while( q->Param() > s && k < n )
		{
			double t = ( s - r ) / ( q->Param() - r );
			Vec2 X = Interpolate( p->Coord(), q->Coord(), t );
			if( t > epsilon && ( 1.0 - t ) > epsilon ) P << X;
			P << Multiline::Jump;
			P << X;
			s = Abs( split[++k] * M.Length() ); 		  
		}
		P << q->Coord();
		r = q->Param();
		p = q;
	}
	return P;
}

Multiline *Explode( const Multiline &M )
{
	// Explode returns an array of single-component multilines created
	// from a single multiline (typically with multiple components).
	// The size of the array is the number of components, Comps().
	// Thie multiline M is not altered.
	Multiline P( M );
	Multiline *list = new Multiline[ M.Comps() ];
	for( int i = 0; i < M.Comps(); i++ ) P.Pop( list[i] );
	return list;
}

Multiline CopyComponent( const Multiline &M, int index ) 
{
	Multiline P;
	Pnode *first = M.CompPtr( index );
	for( Pnode *n = first; n != NULL; n = n->Next() )
	{
		if( n->Jump() && n != first ) break;
		P << *n;
	}
	return P;
}

Multiline Dashed( const Multiline &M, double on, double off, double shift )
{
	// Make a dashed version of the multiline.	"on" is the lenght of the dashes,
	// "off" is the length of the spaces, and "shift" controls where the pattern
	// starts.	All segments of the multiline are processed.
	//
	Multiline P( M ), Q, D;
	
	while( P.Comps() >= 1 )
	{
		P.Pop( Q );  // Process the next segment.
		if( Q.Length() < on + off + shift ) 
		{
			D += Q;
			continue;
		}
		double u 	= ( shift - on - off ) / Q.Length();
		double u_on	= on  / Q.Length();
		double u_off = off / Q.Length();
		
		Pnode *n = Q.First();
		double s  = n->Param() / Q.Length();
		
		while( u < 1.0 )
		{
			if( u >= 0.0 ) D << Q( u ); 	   
			while( s <= u + u_on && n != NULL )
			{
				if( s > u ) D << Q( s );
				n = n->Next();
				if( n != NULL ) s = n->Param() / Q.Length();
			}
			u += u_on;
			if( u > 0.0 ) D << Q( u );		  
			D << Multiline::Jump;
			u += u_off; 
		}
	}
	
	D.SetWeight( M.Weight() );
	return D;
}

double Distance( const Multiline &A, const Multiline &B, double &param_A, double &param_B )
{
	double param;
	double dist2 = Infinity;
	Vec2  X;
	
	for( Pnode *p = A.First(); p != NULL; p = p->Next() )
	{
		X = B.Closest( *p, param );
		double temp = LenSqr( *p - X );
		if( temp < dist2 )
		{
			dist2 = temp;
			param_A = p->Param();
			param_B = param;
		}
	}
	
	for( Pnode *q = B.First(); q != NULL; q = q->Next() )
	{
		X = A.Closest( *q, param );
		double temp = LenSqr( *q - X );
		if( temp < dist2 )
		{
			dist2 = temp;
			param_A = param;
			param_B = q->Param();
		}
	}
	return sqrt( dist2 );
}

double SignedArea( const Multiline &pgon ) 
{
	double area = 0.0;
	if( pgon.Points() < 3 ) return 0.0;
	const Pnode *u = pgon.Last();
	const Pnode *v = pgon.First();
	while( v != NULL )
	{
		area += v->Coord() ^ u->Coord();
		u = v;
		v = v->Next();
	}
	return 0.5 * area;
}

Multiline SpliceAtNearestPoint( const Multiline &A, const Multiline &B )
{
	// Treat A and B as simple closed polygons, and join them with a
	// a single cut line, maintaining a consistent clock-sense in the
	// resulting polygon.  This can be used for changing polygons with
	// holes into simple polygons (for the benefit of simple renderers
	// like Tcl).
	Multiline C( B );
	
	// Ensure that A and C have opposite orientations if they are nested,
	// and the same orientations otherwise.  Do this by reversing the
	// orientation of C if necessary.
	double area_A = SignedArea( A );
	double area_C = SignedArea( C );
	int   nested = A.Inside( C.Start() ) || C.Inside( A.Start() );
	
	if( (  nested && area_A * area_C > 0.0 ) ||
		( !nested && area_A * area_C < 0.0 ) ) C.Reverse(); 
	
	Multiline Q, A1, A2, C1, C2;
	double s, t;
	// Find the closest points on the polygons and the arclength
	// parameters of these points in their respective polygons.
	Distance( A, C, s, t );
	
	// Split each polygon into two pieces at their points of
	// closest approach.
	Split( A, s, A1, A2 );
	Split( C, t, C1, C2 );
	
	// Finally, put the four pieces together into a single polygon in
	// such a way that all interior points of the original (nested)
	// polygons are still interior points of the resulting simple polygon.
	Q << A1;
	Q << C2;
	Q << C1;
	Q << A2;
	
	return Q;
}

// The following strings are used for reading and writing clear-text encodings
// of multilines.  By sharing the same formatting strings, the read and write
// methods are guaranteed to agree on the exact syntax.

static const char
*header   = "Multiline: length = %f, points = %d, components = %d",
*line_out = "  (%5.2f, %5.2f) %7.3f %d",
*line_in  = "  (   %f,    %f)    %f %d";

ostream &operator<<( ostream &out, const Multiline &M )
{
	// Write an ASCII representation of a multiline to the stream in exactly
	// the form expected by the "Read" function.  All time stamps and jumps
	// are preserved.

	// Write the sequence of points (with parameters) to a file.
	out << endl << "Multiline: length = " << M.Length() << ", points = " << M.Points()
		<< ", components = " << M.Comps() << endl;
	for( Pnode *n = M.First(); n != NULL; n = n->Next() )
	{
		if( n->Jump() ) out << endl;
		out << "  (" << n->Coord().X() << ", " << n->Coord().Y() << ") " << n->Param() << " " << n->Time() << endl;
	}
	out << endl;
	out.flush();

	return out;
}

istream &operator>>( istream &in, Multiline &M )
{
	char in_buffer[256];
	// Read a multiline back from its ASCII representation, as written
	// by the >> operator, for example.  All time stamps and jumps
	// are preserved.
	M.Clear();
	
	// Read in the header line, which gives the total length, number of
	// vertices, and number of components.
	float len;
	int npts, comps;
	skipWhite( in );
	in.getline( in_buffer, sizeof(in_buffer) );
	if( !in || sscanf( in_buffer, header, &len, &npts, &comps ) != 3 )
	{
		cerr << "ERROR (Multiline::Read) first line of multiline is bad" << endl;
		return in;
	}
	
	// Read in the rest of the vertices, one line at a time.  The arclength parameter
	// is used only to determine where the jumps are, so only zeros are significant.
	// Intermediate parameters are recomputed.	This guarantees that the arclength
	// parameter is correct, since the multiline being read in may have been edited by 
	// hand (to add or delete points, for example).  The total length is also recomputed
	// for the same reason.
	for( int i = 1; i <= npts; i++ )
	{
		float x, y, param;
		long  time;
		skipWhite( in );
		in.getline( in_buffer, sizeof(in_buffer) );
		int num_read = sscanf( in_buffer, line_in, &x, &y, &param, &time );
		if( num_read != 4 )
		{
			cerr << "ERROR (Multiline::Read) field " << num_read + 1 
				<< " of point " << i << " could not be read:";
			if( num_read >= 1 ) cerr << "  x = " << x;
			if( num_read >= 2 ) cerr << ", y = " << y;
			if( num_read >= 3 ) cerr << ", p = " << param;
			cerr << ", <bad field>" << endl;
			return in;
		}
		if( param == 0.0 ) M << Multiline::Jump;
		M << TimedVec2( x, y, time );
	}
	return in;
}


// Refine the multiline P by inserting new points along every segment that
// is longer than "eps".  The resulting multiline that is returned is
// the same geometrically, but no segment will be longer than "eps".
//
Multiline Resample( const Multiline &P, double eps )
{
	Multiline Q;
	Pnode *p = NULL;
	Pnode *q = P.First();
	
	if( q == NULL ) return P;
	// Insert the starting point into the new Multiline.
	Q << q->Coord();
	
	while( q->Next() != NULL )
	{
		p = q;
		q = q->Next();
		if( q->Jump() )
		{
			// Don't insert points in the breaks between segments.
			Q << Multiline::Jump;
		}
		else
		{
			// Compute how many points to insert, which are to be
			// interpolated at evenly-spaced intervals.
			int steps = (int)( TwoNorm( q->Coord() - p->Coord() ) / eps );
			for( int i = 1; i <= steps; i++ )
			{
				double t = double(i) / ( steps + 1 );
				Q << Interpolate( p->Coord(), q->Coord(), t );
			}
		}
		Q << q->Coord();
	}
	return Q;
}

