/***************************************************************************
* Name.C																   *
*																		   *
* Defines "NameTableItem", which is the base class for objects that can be *
* inserted into a "NameTable" object.									   *
*																		   *
*	Changes:															   *
*																		   *
*		08/06/99	arvo	  Added secondary qualifiers to names.		   *
*		08/10/98	arvo	  Moved from font directory to util.		   *
*		06/06/98	arvo	  Initial coding.							   *
*																		   *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo										   *
*																		   *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html	   *
*																		   *
* This program is distributed in the hope that it will be useful, but	   *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more	   *
* details.																   *
*																		   *
***************************************************************************/
#include "stdafx.h"

#include <stdlib.h>
#include <string.h>
//#include <iostream.h>
#include "../util/Char.h"
#include "../util/Name.h"

static int factor[] = { 13, 67, 107, 171, 211, 7, 11 };

static long Code( const char *name )
{
	int i = 0;	  
	int k = sizeof(factor) / sizeof(int);
	long code = 0;
	// Compute the hash code based on the primary key only.
	while( *name != NullChar )
	{
		code += ( *name++ ) * factor[ i++ ];
		if( i == k ) i = 0;
	}
	return code > 0 ? code : -code;
}

NameTableItem::NameTableItem( const char *n, const char *q )
{
	static char *blank = "";
	name  = ( n == NULL ) ? blank : strdup( n );
	qual  = ( q == NULL ) ? blank : strdup( q );
	code  = Code( n );
	next  = NULL;
	back  = NULL;
	owner = NULL;
}

NameTable::NameTable()
{
	Clear();
}

NameTable::~NameTable()
{
	Clear();
}

NameTableItem *NameTable::Lookup( const char *name, const char *qual ) const
{
	long c = Code( name );
	NameTableItem *item = table[ c % NameHashTableSize ];
	while( item != NULL )
	{
		// Use the long version of the hash code as a first test
		// for equality, before doing a string comparison.
		if( item->code == c && Match( item->name, name ) ) 
		{
			if( qual == NULL || Match( item->qual, qual ) ) break;
		}
		item = item->next;
	}
	return item;
}

void NameTable::Clear()
{
	for( register int i = 0; i < NameHashTableSize; i++ ) table[i] = NULL;
	numentries = 0;
}

void NameTable::DeleteAll()
{
	for( register int i = 0; i < NameHashTableSize; i++ ) 
	{
		NameTableItem *item = table[i];
		while( item != NULL )
		{
			NameTableItem *next = item->next;
			delete item;
			item = next;
		}
		table[i] = NULL;
	}
	numentries = 0;
}

NameTableItem *NameTable::operator>>( NameTableItem *item )
{
	if( item == NULL || item->owner != this ) return NULL;
	if( item->back != NULL ) 
		item->back->next = item->next;
	else
	{
		int k = ( item->code % NameHashTableSize );
		table[k] = item->next;
	}
	if( item->next != NULL ) item->next->back = item->back;
	item->next	= NULL;    
	item->back	= NULL;    
	item->owner = NULL;    
	numentries--;
	return item;
}

NameTable &NameTable::operator<<( NameTableItem *item )
{
	if( item == NULL ) return *this;
	// If the item is already owned by a name table, check to see
	// which table it is, and act accordingly (since an item can only
	// a member of one table at a time).
	if( item->owner != NULL ) 
	{
		if( item->owner == this ) return *this;  // It's already in the name table.
		*(item->owner) >> item; 				 // Remove it from the other table it's in.
	}
	int k = ( item->code % NameHashTableSize );
	item->next	= table[k];
	item->back	= NULL;
	item->owner = this;
	if( table[k] != NULL ) table[k]->back = item;
	table[k] = item;
	numentries++;
	return *this;
}

// Return a pointer to the first item in the hash table.
NameTableItem *NameTable::First() const
{
	NameTableItem *item = NULL;
	for( int i = 0; i < NameHashTableSize; i++ ) 
	{
		if( table[i] != NULL )
		{
			item = table[i];
			break;
		}
	}
	return item;
}

// Return a pointer to the item in the hash table that
// follows a given item.  Note that the order is determined
// primarily by the hash code, and secondarily by the order 
// of item entry (when there are hash table collisions).
NameTableItem *NameTable::Next( const NameTableItem *item ) const
{
	NameTableItem *next = NULL;
	if( item == NULL )
	{
		// This is the end of the list.
	}
	else if( item->next != NULL )
	{
		// Advance to the next item in this hash table row.
		next = item->next;
	}
	else 
	{
		// We are at the end of a row in the hash table.  Get
		// the row number from the hash code and start searching
		// the table beginning with the next row.
		int index = item->code % NameHashTableSize;
		for( int i = index + 1; i < NameHashTableSize; i++ ) 
		{
			if( table[i] != NULL ) 
			{
				next = table[i];
				break;
			}
		}
	}
	
	return next;
}
