/***************************************************************************
* Trans2.h                                                                 *
*                                                                          *
* Two-dimensional transformations with uniform (isotropic) scaling.        *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       07/04/98    arvo    First working version.                         *
*       04/30/98    arvo    Initial coding.                                *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#ifndef __TRANS2_INCLUDED__
#define __TRANS2_INCLUDED__

#include "../util/Vec2.h"
#include "../util/Box2.h"

// The Trans2 class encapsulates a 2D transformation that consists
// of a translation and a uniform scale.  It simply extends a 2D vector
// to include a single scalar value.

class Trans2 : public Vec2 
    {

    public:
        Trans2();
        Trans2( const Vec2 &O, double s = 1.0 );
        Trans2( const Trans2 &t ) { *this = t; }
       ~Trans2() {}
        inline const Vec2   &Origin() const { return *this; }
        inline       Vec2   &Origin()       { return *this; }
        inline       double  Scale () const { return scale; }
        inline       double &Scale ()       { return scale; }
        inline Trans2  Inverse() const;
        inline Trans2 &operator= ( const Trans2 & );
        inline Trans2 &operator= ( const Vec2   & );
        inline Trans2 &operator*=( const Trans2 & );
        inline Trans2 &operator/=( const Trans2 & );
        inline Trans2 &operator+=( const Vec2   & );
        inline Trans2 &operator-=( const Vec2   & );
        inline Trans2 &operator*=( double           );
        inline Trans2 &operator/=( double           );
        inline Trans2  operator* ( const Trans2 & ) const;
        inline Trans2  operator/ ( const Trans2 & ) const;
        inline Trans2  operator+ ( const Vec2   & ) const;
        inline Trans2  operator- ( const Vec2   & ) const;

        const char *ToString( char *buff = NULL ) const;
        Trans2 &FromString( const char *str );

    public: // Static constants and functions.
        static const Trans2 Null;
        static const Trans2 Ident;
        static       Trans2 Scale( double s );

    private:
        double scale;
    };

inline Trans2::Trans2()
    { 
    Origin() = 0.0;
    Scale () = 1.0; 
    }

inline Trans2::Trans2( const Vec2 &O, double s ) 
    { 
    Origin() = O;
    Scale () = s;
    }

inline Trans2 Trans2::Inverse() const
    {
    return Trans2( -Origin() / Scale(), 1.0 / Scale() );
    }

inline Trans2 &Trans2::operator=( const Trans2 &t )
    {
    Origin() = t.Origin();
    Scale () = t.Scale ();
    return *this;
    }

inline Trans2 &Trans2::operator=( const Vec2 &O )
    {
    Origin() = O;
    Scale () = 1.0;
    return *this;
    }

inline Trans2 &Trans2::operator*=( const Trans2 &t )
    {
    Origin() += t.Origin() * scale;
    Scale () *= t.Scale ();
    return *this;
    }

inline Trans2 &Trans2::operator/=( const Trans2 &t )
    {
    double s   = Scale() / t.Scale();
    Scale ()  = s;
    Origin() -= s * t.Origin();
    return *this;
    }

inline Trans2 Trans2::operator/( const Trans2 &t ) const
    {
    double s = Scale() / t.Scale();
    return Trans2( Origin() - s * t.Origin(), s );
    }

inline Trans2 &Trans2::operator+=( const Vec2 &V )
    {
    Origin() += V;
    return *this;
    }

inline Trans2 &Trans2::operator-=( const Vec2 &V )
    {
    Origin() -= V;
    return *this;
    }

inline Trans2 &Trans2::operator*=( double s )
    {
    Scale() *= s;
    return *this;
    }

inline Trans2 &Trans2::operator/=( double s )
    {
    Scale() /= s;
    return *this;
    }

inline Trans2 Trans2::operator*( const Trans2 &t ) const
    {
    return Trans2( 
      Origin() + t.Origin() * Scale(),
      Scale () * t.Scale ()
      );
    }

inline Trans2 Trans2::operator+( const Vec2 &V ) const
    {
    return Trans2( Origin() + V, Scale() );
    }

inline Trans2 Trans2::operator-( const Vec2 &V ) const
    {
    return Trans2( Origin() - V, Scale() );
    }


// Non-member functions involving the Trans2 class.

inline Vec2 operator*( const Vec2 &vec, const Trans2 &trans )
    {
    return trans.Scale() * vec + trans.Origin();
    }

inline Vec2 &operator*=( Vec2 &vec, const Trans2 &trans )
    {
    vec = trans.Scale() * vec + trans.Origin();
    return vec;
    }

inline Vec2 operator*( const Trans2 &trans, const Vec2 &vec )
    {
    return trans.Scale() * vec + trans.Origin();
    }

inline Vec2 operator/( const Vec2 &vec, const Trans2 &trans )
    {
    return ( vec - trans.Origin() ) / trans.Scale();
    }

inline Box2d operator*( const Box2d &box, const Trans2 &trans )
    {
    return Box2d( trans * box.Min(), trans * box.Max() );
    }

inline Box2d &operator*=( Box2d &box, const Trans2 &trans )
    {
    return box.Set( trans * box.Min(), trans * box.Max() );
    }

inline Box2d operator*( const Trans2 &trans, const Box2d &box )
    {
    return Box2d( trans * box.Min(), trans * box.Max() );
    }

// Return the box transformed by the inverse of the given Trans2.
inline Box2d operator/( const Box2d &box, const Trans2 &trans )
    {
    return box * trans.Inverse();
    }

// Linearly interpolate between transformations A and B according to
// the parameter t, which ranges from 0 to 1.
inline Trans2 Interpolate( const Trans2 &A, const Trans2 &B, double t )
    {
    double s = 1.0 - t;
    return Trans2(
        s * A.Origin() + t * B.Origin(),
        s * A.Scale () + t * B.Scale ()
        );
    }

// The 2-norm of the greatest distance between corresponding points on
// the box under the two transformations.  This can be used as a measure
// of the "distance" between the two transformations.
extern double dist( 
    const Box2d   &,
    const Trans2 &, 
    const Trans2 & 
    );

// Return a 2-D isotropic transformation that maps one 2-D
// box into another, as closely as possible.  By default, the
// Y-extent of the transformed box will exactly match that of
// the target box.  If the third argument is 1, then the X-extent
// will be matched exactly.
extern Trans2 Box_to_Box( 
    const Box2d &from, // The box to transform.
    const Box2d &to,   // The target box.
    int match_x = 0    // 1 ==> match X, 0 ==> match Y.
    );

extern ostream &operator<<( ostream &out, const Trans2 & );

#endif

